#include <vector>
#include <cassert>
#include <fstream>
#include <iostream>
#include <map>
#include <unordered_map>
#include <list>
#include <graph.hpp>
#include <vector>
#include <cmath>
#include <iomanip>
#include <boost/graph/topological_sort.hpp>
#include <boost/graph/connected_components.hpp>
#include <ext/pb_ds/assoc_container.hpp>
#include "tables.hpp"

using namespace std;
using namespace __gnu_pbds;

size_t (*citefunc)(const list<dv_t> &verts, const instance &inst);

instance read_instance(const char *gname, const char *dname) {
  instance res;
  trie<string, dv_t> Vmap;
  
  string line;

  size_t a_, b_;
  
  ifstream ifs(dname);
  while(getline(ifs,line)) {
    if(line.empty()) continue;
    istringstream iss(line);
    string a,b;
    iss >> a >> b;

    if(Vmap.find(a) == Vmap.end()) {
      a_ = add_vertex(res.d);
      if(gname) assert(a_ == add_vertex(res.g));
      Vmap[a] = a_;
    } else a_ = Vmap[a];
    if(Vmap.find(b) == Vmap.end()) {
      b_ = add_vertex(res.d);
      if(gname) assert(b_ == add_vertex(res.g));
      Vmap[b] = b_;
    } else b_ = Vmap[b];
    
    add_edge(b_, a_, res.d);
  }
  
  cout << "Tot: " << setw(4) << num_vertices(res.d) << " articles and " << setw(5) << num_edges(res.d) << " citations " << endl;

  res.to_name.resize(num_vertices(res.d));

  for(auto &it : Vmap) res.to_name[it.second] = it.first;

  if(!gname) return res;

  ifstream ifs_(gname);
  while(getline(ifs_,line)) {
    if(line.empty()) continue;
    istringstream iss(line);
    string a,b;
    iss >> a >> b;

    if(Vmap.find(a) == Vmap.end()) {
      cerr << "W: discarding uncited article " << a << " from compatibility graph" << endl;
      continue;
    }
    if(Vmap.find(b) == Vmap.end()) {
      cerr << "W: discarding uncited article " << b << " from compatibility graph" << endl;
      continue;
    }
    a_ = Vmap[a];
    b_ = Vmap[b];
    
    if(cites(res, a_) <= 0 or cites(res, b_) <= 0) continue;
    
    add_edge(b_, a_, res.g);
  }

  cout << "Read " << setw(4) << num_edges(res.g) << " compatibilities" << endl;

  //  vector<int> component(num_vertices(res.g));
  int num = connected_components(res.g, get(&vertinfo::annotation, res.g));
  res.comps.resize(num);

  uv_iter vs, ve;
  for(tie(vs,ve) = vertices(res.g); vs != ve; ++vs)
    ++res.comps[res.g[*vs].annotation].c;
  
  for(int i = 0; i < num; ++i) {
    if(res.comps[i].c > 1)
      cout << setw(9) << res.comps[i].c << " vertices in component " << setw(4) << i << endl;
    res.comps[i].to_idx.resize(res.comps[i].c);
    res.comps[i].c = 0;
  }

  for(tie(vs,ve) = vertices(res.g); vs != ve; ++vs)
    res.comps[res.g[*vs].annotation].to_idx[res.comps[res.g[*vs].annotation].c++] = *vs;

  return res;
}

size_t ucites(const list<dv_t> &verts, const instance &inst) {
  vector<int> cites_me(num_vertices(inst.d), 0);

  in_iter e0, en; 
  for(auto &v : verts)
    for(tie(e0, en) = in_edges(v, inst.d); e0 != en; ++e0)
      cites_me[source(*e0, inst.d)] = 1;

  size_t res = 0;
  for(auto &cite : cites_me) res += cite;

  return res;
}

size_t scites(const list<dv_t> &verts, const instance &inst) {
  size_t res = 0;

  for(auto &v : verts)
    res += cites(inst, v);

  return res;
}

size_t cites(const instance &inst, const dv_t& v){
  return in_degree(v, inst.d);
}

bool is_clique(const list<uv_t> &verts, instance &inst) {
  for(auto &v : verts) inst.g[v].annotation = 0;

  e_iter e0, en;
  for(auto &v : verts) 
    for(tie(e0, en) = out_edges(v, inst.g); e0 != en; ++e0)
      ++inst.g[target(*e0, inst.g)].annotation;

  int s = verts.size();

  for(auto &v : verts) {
    assert(inst.g[v].annotation < s);
    if(inst.g[v].annotation < s - 1) return false;
  }

  return true;
}

size_t hindex(const instance &inst, unsigned purged) {
  map<unsigned,unsigned> num_articles_with_cites;

  dv_iter v0, vn;
  tie(v0, vn) = vertices(inst.d);
  for(; v0 != vn; ++v0)
    ++num_articles_with_cites[cites(inst, *v0)];
  
  size_t arts_seen(purged);
  size_t h(purged);

  cout << setw(9) << arts_seen << " articles with    ∞ citations" << endl;
  
  for(auto i = num_articles_with_cites.rbegin(); 
      i != num_articles_with_cites.rend(); ++i) {
    cout << setw(9) << i-> second << " articles with " << setw(4) << i->first << " citations" << endl;
    for(unsigned j = 1; j <= i->second; ++j)
      if(++arts_seen <= i->first) ++h;
  }
  
  return h;
}

bool can_reach_hindex(instance &inst, unsigned h, unsigned purged) {
  size_t articles = 0;
  for(size_t C = 0; C < inst.comps.size(); ++C) {
    //    cout << "trying component " << C << ": ";
    size_t res = above_h_articles(inst, C, h);
    //    cout << res << " articles with at least " << h << " citations" << endl;
    articles += res;

    if(purged + articles >= h) return true;
  }
  return false;
}

size_t preprocess_for_k_merges(instance &inst, unsigned h, unsigned k) {
  assert(num_vertices(inst.g) == 0);

  size_t thrown_out = 0;

  multimap<unsigned, dv_t> arts_by_cites;

  dv_iter v0, vn;
  tie(v0, vn) = vertices(inst.d);

  for(; v0 != vn; ++v0) {
    unsigned c = cites(inst, *v0);
    if(c == 0) continue;
    else if(c < h) arts_by_cites.insert(make_pair(c, *v0));
    else ++thrown_out;
  }

  inst.comps.resize(1);
  inst.comps[0].c = 0;
  inst.comps[0].to_idx.resize(2*k);

  for(auto i = arts_by_cites.rbegin(); i != arts_by_cites.rend(); ++i) {
    inst.comps[0].to_idx[inst.comps[0].c++] = i->second;
    if(inst.comps[0].c >= 2*k) break;
  }
  
  cout << "Preprocessed for reaching H-index " << h << " with " << k << " merges:" << endl;

  cout << "   " << thrown_out << " articles with sufficiently high citation count" << endl;
  cout << "   " << inst.comps[0].c << " articles remaining" << endl;

  return thrown_out;
}

bool can_reach_hindex_k(instance &inst, unsigned h, unsigned k, unsigned purged) {
  size_t to = preprocess_for_k_merges(inst, h, k);
  size_t found = above_h_articles_k(inst, h, k);

  cout << "   created " << found << " additional sufficient articles" << endl;

  return(h <= to + found + purged);
}

size_t hindex_ubo(const instance &inst, unsigned purged) {
  // To get H-index h, we have to create (h-purge) articles with at
  // least h cites.  That is, the number c of input citations has to
  // satisfy c >= (h-purge)h, equivalently 0 >= h^2-h*purge-c, that is
  // h <= 1/2 (sqrt(4c+p^2)+p)

  return floor(0.5*(sqrt(4.0*num_edges(inst.d)+pow(purged, 2.0))+purged));
}

